from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from letta_client import AgentState, LettaMessageUnion

from letta_evals.models import GradeResult, Sample


class Grader(ABC):
    """Base interface for graders."""

    @property
    @abstractmethod
    def requires_agent_state(self) -> bool:
        """Whether this grader requires agent_state for extraction."""
        pass

    @abstractmethod
    async def grade(
        self, sample: Sample, trajectory: List[List[LettaMessageUnion]], agent_state: Optional[AgentState] = None
    ) -> Tuple[GradeResult, str]:
        """Grade a trajectory and return the result and extracted submission."""
        pass



# import built-in graders to trigger registration via decorators
from letta_evals.graders.builtin import ascii_printable_only, contains, exact_match

__all__ = ["contains", "exact_match", "ascii_printable_only"]



import json
import os
from pathlib import Path
from typing import List, Optional, Tuple

from dotenv import load_dotenv
from letta_client import AgentState, LettaMessageUnion
from openai import AsyncOpenAI

from letta_evals.extractors import extractor_requires_agent_state, get_extractor
from letta_evals.graders.base import Grader
from letta_evals.models import GradeResult, Sample
from letta_evals.types import LLMProvider

load_dotenv()


class RubricGrader(Grader):
    """Grader that uses an LLM judge with custom rubric prompts."""

    def __init__(
        self,
        prompt: str,
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        provider: LLMProvider = LLMProvider.OPENAI,
        max_retries: int = 5,
        timeout: float = 120.0,
        extractor: str = "last_assistant",
        extractor_config: Optional[dict] = None,
        base_dir: Optional[Path] = None,
    ):
        self.prompt = prompt
        self.model = model
        self.temperature = temperature
        self.provider = provider
        self.extractor_name = extractor
        self.base_dir = base_dir
        self.extractor = get_extractor(extractor, extractor_config, base_dir=base_dir)
        self._requires_agent_state = extractor_requires_agent_state(extractor, base_dir=base_dir)

        if provider == LLMProvider.OPENAI:
            api_key = os.getenv("OPENAI_API_KEY")
            if not api_key:
                raise ValueError("OPENAI_API_KEY not found in environment variables")

            client_kwargs = {
                "api_key": api_key,
                "max_retries": max_retries,
                "timeout": timeout,
            }

            base_url = os.getenv("OPENAI_BASE_URL")
            if base_url:
                client_kwargs["base_url"] = base_url

            self.client = AsyncOpenAI(**client_kwargs)
        else:
            raise ValueError(f"Unsupported provider: {provider}")

    @property
    def requires_agent_state(self) -> bool:
        """Whether this grader's extractor requires agent_state."""
        return self._requires_agent_state

    async def grade(
        self, sample: Sample, trajectory: List[List[LettaMessageUnion]], agent_state: Optional[AgentState] = None
    ) -> Tuple[GradeResult, str]:
        """Grade using LLM judge with rubric."""
        submission = self.extractor(trajectory, agent_state=agent_state)

        judge_prompt = self._build_judge_prompt(sample, submission)

        temperature = self.temperature
        if (
            self.model.startswith("o1") or self.model.startswith("o3") or "gpt-5" in self.model.lower()
        ) and temperature == 0.0:
            temperature = 1.0

        try:
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": self._get_system_prompt()},
                    {"role": "user", "content": judge_prompt},
                ],
                temperature=temperature,
                response_format={"type": "json_object"},
            )

            result_json = json.loads(response.choices[0].message.content)

            score = result_json.get("score")
            if score is None:
                raise ValueError("Model did not return a score")

            score = float(score)
            score = max(0.0, min(1.0, score))

            return GradeResult(
                score=score,
                rationale=result_json.get("rationale", ""),
                metadata={"model": self.model, "usage": response.usage.model_dump() if response.usage else None},
            ), submission

        except Exception as e:
            return GradeResult(
                score=0.0, rationale=f"Error during grading: {str(e)}", metadata={"error": str(e)}
            ), submission

    def _get_system_prompt(self) -> str:
        """Get the system prompt for the judge."""
        return """You are an evaluation judge. You will be given:
1. A rubric describing evaluation criteria
2. An input/question
3. A submission to evaluate

Evaluate the submission according to the rubric and return a JSON response with:
{
    "score": (REQUIRED: a decimal number between 0.0 and 1.0 inclusive),
    "rationale": "explanation of your grading decision"
}

IMPORTANT:
- The score MUST be a number between 0.0 and 1.0 (inclusive)
- 0.0 means complete failure, 1.0 means perfect
- Use decimal values for partial credit (e.g., 0.25, 0.5, 0.75)
- Be objective and follow the rubric strictly"""

    def _build_judge_prompt(self, sample: Sample, submission: str) -> str:
        """Build the prompt for the judge."""
        prompt = self.prompt

        prompt = prompt.replace("{input}", str(sample.input))
        prompt = prompt.replace("{submission}", submission)

        if sample.ground_truth:
            prompt = prompt.replace("{ground_truth}", sample.ground_truth)

        parts = [
            "## Rubric",
            prompt,
            "",
            "## Input",
            str(sample.input),
        ]

        if sample.ground_truth:
            parts.extend(["", "## Ground Truth Answer", sample.ground_truth])

        parts.extend(
            [
                "",
                "## Submission to Evaluate",
                submission,
                "",
                "Please evaluate the submission according to the rubric and return your judgment in JSON format.",
            ]
        )

        return "\n".join(parts)



from pathlib import Path
from typing import Optional

from letta_client import AsyncLetta, LlmConfig, MessageCreate

from letta_evals.models import Sample, TargetResult
from letta_evals.targets.base import Target
from letta_evals.utils import load_object
from letta_evals.visualization.base import ProgressCallback


class AgentTarget(Target):
    """Letta agent target for evaluation."""

    def __init__(
        self,
        client: AsyncLetta,
        agent_id: str = None,
        agent_file: Path = None,
        agent_script: str = None,
        base_dir: Path = None,
        llm_config: Optional[LlmConfig] = None,
        model_handle: Optional[str] = None,
    ):
        self.client = client
        self.agent_id = agent_id
        self.agent_file = agent_file
        self.agent_script = agent_script
        self.base_dir = base_dir or Path.cwd()
        self.llm_config = llm_config
        self.model_handle = model_handle

    async def run(
        self,
        sample: Sample,
        progress_callback: Optional[ProgressCallback] = None,
        project_id: Optional[str] = None,
        retrieve_agent_state: bool = False,
    ) -> TargetResult:
        """Run the agent on a sample."""
        agent_id = self.agent_id

        if self.agent_file:
            with open(self.agent_file, "rb") as f:
                resp = await self.client.agents.import_file(
                    file=f, append_copy_suffix=False, override_existing_tools=False, project_id=project_id
                )
                if len(resp.agent_ids) > 1:
                    raise RuntimeError(
                        f"Expected single agent from .af file, got {len(resp.agent_ids)} agents. We don't support multi-agent evals yet."
                    )

                agent_id = resp.agent_ids[0]

        elif self.agent_script:
            agent_factory_func = load_object(self.agent_script, self.base_dir)
            agent_id = await agent_factory_func(self.client, sample)

        if self.llm_config and agent_id:
            await self.client.agents.modify(agent_id=agent_id, llm_config=self.llm_config)
        elif self.model_handle and agent_id:
            await self.client.agents.modify(agent_id=agent_id, model=self.model_handle)

        agent = await self.client.agents.retrieve(agent_id=agent_id, include_relationships=[])
        if self.llm_config:
            model_name = self.llm_config.model
        elif self.model_handle:
            model_name = self.model_handle
        else:
            model_name = agent.llm_config.model

        if progress_callback and (self.agent_file or self.agent_script):
            await progress_callback.agent_loading(sample.id, model_name=model_name)

        trajectory = []
        usage_stats: list[dict] = []

        inputs = sample.input if isinstance(sample.input, list) else [sample.input]
        total_messages = len(inputs)

        for i, input_msg in enumerate(inputs):
            if progress_callback:
                await progress_callback.message_sending(sample.id, i + 1, total_messages, model_name=model_name)

            stream = self.client.agents.messages.create_stream(
                agent_id=agent_id,
                messages=[MessageCreate(role="user", content=str(input_msg))],
                stream_tokens=True,
            )

            run_id = None
            async for chunk in stream:
                # derive run_id from very first chunk, all should have the same
                if not run_id:
                    run_id = chunk.run_id

                # handle usage statistics in a streaming fashion
                if hasattr(chunk, "message_type"):
                    if chunk.message_type == "usage_statistics":
                        # best-effort convert to JSON-serializable dict
                        usage_rec = None
                        if hasattr(chunk, "model_dump") and callable(getattr(chunk, "model_dump")):
                            try:
                                usage_rec = chunk.model_dump()
                            except Exception:
                                usage_rec = None
                        if usage_rec is None and hasattr(chunk, "dict") and callable(getattr(chunk, "dict")):
                            try:
                                usage_rec = chunk.dict()  # type: ignore[attr-defined]
                            except Exception:
                                usage_rec = None
                        if usage_rec is None and hasattr(chunk, "__dict__"):
                            try:
                                usage_rec = dict(chunk.__dict__)
                            except Exception:
                                usage_rec = None
                        if usage_rec is None:
                            # final fallback to string
                            usage_rec = {"raw": str(chunk)}
                        usage_stats.append(usage_rec)
                        continue

            if not run_id:
                raise RuntimeError("Unexpected error: no run ID was returned from streaming chunks.")

            # TODO: Set limit here potentially, this is capped to 100
            messages = await self.client.runs.messages.list(run_id=run_id)
            trajectory.append(messages)

        # conditionally retrieve final agent state if needed (includes memory blocks)
        final_agent_state = None
        if retrieve_agent_state:
            final_agent_state = await self.client.agents.retrieve(agent_id=agent_id, include_relationships=[])

        return TargetResult(
            trajectory=trajectory,
            agent_id=agent_id,
            model_name=model_name,
            agent_usage=usage_stats,
            agent_state=final_agent_state,
        )



import re
from typing import List, Optional

from letta_client import AgentState, AssistantMessage, LettaMessageUnion, ToolCallMessage, ToolReturnMessage

from letta_evals.decorators import extractor
from letta_evals.extractors.utils import (
    flatten_content,
    get_assistant_messages,
    get_last_turn_messages,
)


@extractor
def last_assistant(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract the last assistant message content."""
    messages = get_assistant_messages(trajectory)
    if not messages:
        return ""
    return flatten_content(messages[-1].content)


@extractor
def first_assistant(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract the first assistant message content."""
    messages = get_assistant_messages(trajectory)
    if not messages:
        return ""
    return flatten_content(messages[0].content)


@extractor
def all_assistant(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Concatenate all assistant messages."""
    messages = get_assistant_messages(trajectory)
    contents = [flatten_content(msg.content) for msg in messages]
    separator = config.get("separator", "\n")
    return separator.join(contents)


@extractor
def last_turn(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract all assistant messages from the last turn."""
    messages = get_last_turn_messages(trajectory, AssistantMessage)
    contents = [flatten_content(msg.content) for msg in messages]
    separator = config.get("separator", "\n")
    return separator.join(contents)


@extractor
def pattern(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract content matching a regex pattern."""
    pattern_str = config["pattern"]
    group = config.get("group", 0)
    search_all = config.get("search_all", False)

    messages = get_assistant_messages(trajectory)

    for msg in reversed(messages):
        content = flatten_content(msg.content)
        if search_all:
            matches = re.findall(pattern_str, content)
            if matches:
                if isinstance(matches[0], tuple):
                    return " ".join(m[group] for m in matches)
                else:
                    return " ".join(matches)
        else:
            match = re.search(pattern_str, content)
            if match:
                return match.group(group)

    return ""


@extractor
def tool_arguments(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract arguments from specific tool calls."""
    tool_name = config["tool_name"]

    for turn in trajectory:
        for message in turn:
            if isinstance(message, ToolCallMessage):
                if message.tool_call.name == tool_name:
                    return message.tool_call.arguments

    return "{}"


@extractor
def tool_output(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract output from specific tool calls."""
    tool_name = config["tool_name"]

    # find the tool call first
    tool_call_id = None
    for turn in trajectory:
        for message in turn:
            if isinstance(message, ToolCallMessage):
                if message.tool_call.name == tool_name:
                    tool_call_id = message.tool_call.tool_call_id
                    break
            if tool_call_id:
                break
        if tool_call_id:
            break

    # if we found a matching tool call, find its return
    if tool_call_id:
        for turn in trajectory:
            for message in turn:
                if isinstance(message, ToolReturnMessage) and message.tool_call_id == tool_call_id:
                    return message.tool_return

    return ""


@extractor
def after_marker(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
    """Extract content after a specific marker."""
    marker = config["marker"]
    include_marker = config.get("include_marker", False)

    messages = get_assistant_messages(trajectory)

    for msg in reversed(messages):
        content = flatten_content(msg.content)
        idx = content.find(marker)
        if idx >= 0:
            if include_marker:
                return content[idx:].strip()
            else:
                return content[idx + len(marker) :].strip()

    return ""


@extractor
def memory_block(trajectory: List[List[LettaMessageUnion]], config: dict, agent_state: Optional[AgentState]) -> str:
    """Extract content from a specific memory block by label.

    Config:
        block_label (str): The label of the memory block to extract (e.g., "human", "persona")

    Example usage in suite YAML:
        graders:
          human_memory:
            kind: tool
            function: exact_match
            extractor: memory_block
            extractor_config:
              block_label: "human"
    """
    if agent_state is None:
        raise RuntimeError("memory_block extractor requires agent_state, but it was not retrieved")

    block_label = config.get("block_label")
    if not block_label:
        raise ValueError("memory_block extractor requires 'block_label' in config")

    # search for the block with the specified label
    for block in agent_state.memory.blocks:
        if block.label == block_label:
            return block.value

    # block not found - return empty string
    return ""



import json
from pathlib import Path
from typing import List, Optional, Tuple

import anyio

from letta_evals.models import Metrics, RunnerResult, SampleResult


class StreamingWriter:
    """Writes evaluation outputs incrementally to a JSONL file.

    Record schema per line:
      - header:  {"type": "header", "suite": str, "config": {...}}
      - result:  {"type": "result", "result": SampleResult}
      - summary: {"type": "summary", "metrics": Metrics, "gates_passed": bool}
    """

    def __init__(self, output_path: Path, suite_name: str, config: dict):
        self.output_path = Path(output_path)
        self.suite_name = suite_name
        self.config = config
        self.output_path.mkdir(parents=True, exist_ok=True)

    async def initialize(self) -> None:
        # truncate and write header
        header_obj = {"type": "header", "suite": self.suite_name, "config": self.config}
        with open(self.output_path / "header.json", "w", encoding="utf-8") as f:
            json.dump(header_obj, f, indent=2)

    async def append_result(self, result: SampleResult) -> None:
        result_obj = json.loads(result.model_dump_json())
        await self._write_line({"type": "result", "result": result_obj})

    async def write_metrics(self, metrics: Metrics, gates_passed: bool) -> None:
        metrics_obj = json.loads(metrics.model_dump_json())
        summary_obj = {"type": "summary", "metrics": metrics_obj, "gates_passed": gates_passed}
        with open(self.output_path / "summary.json", "w", encoding="utf-8") as f:
            json.dump(summary_obj, f, indent=2)

    async def _write_line(self, obj: dict, truncate: bool = False) -> None:
        mode = "w" if truncate else "a"
        line = json.dumps(obj, ensure_ascii=False) + "\n"

        def _write() -> None:
            with open(self.output_path / "results.jsonl", mode, encoding="utf-8") as f:
                f.write(line)

        await anyio.to_thread.run_sync(_write)


class StreamingReader:
    """Reads a JSONL streaming results directory back into a RunnerResult."""

    @staticmethod
    async def to_runner_result(path: Path) -> RunnerResult:
        def _read() -> Tuple[Optional[str], Optional[dict], List[SampleResult], Optional[Metrics], bool]:
            _suite: Optional[str] = None
            _config: Optional[dict] = None
            _results: List[SampleResult] = []
            _metrics: Optional[Metrics] = None
            _gates: bool = False

            with open(path / "header.json", "r", encoding="utf-8") as f:
                header_obj = json.load(f)
                _suite = header_obj.get("suite")
                _config = header_obj.get("config")

            with open(path / "summary.json", "r", encoding="utf-8") as f:
                summary_obj = json.load(f)
                _metrics = Metrics(**summary_obj.get("metrics"))
                _gates = bool(summary_obj.get("gates_passed", False))

            with open(path / "results.jsonl", "r", encoding="utf-8") as f:
                for raw_line in f:
                    line = raw_line.strip()
                    if not line:
                        continue
                    rec = json.loads(line)
                    _results.append(SampleResult(**rec["result"]))
            return _suite, _config, _results, _metrics, _gates

        suite, config, results, metrics, gates_passed = await anyio.to_thread.run_sync(_read)

        if suite is None or config is None:
            raise ValueError("Results JSONL missing header record")
        if metrics is None:
            raise ValueError("Results JSONL missing summary record")

        return RunnerResult(suite=suite, config=config, results=results, metrics=metrics, gates_passed=gates_passed)

    # No fallback metrics: summary is required in JSONL results.



import sys
from pathlib import Path
from typing import Optional

import anyio
import typer
import yaml
from rich.console import Console
from rich.table import Table

from letta_evals.constants import MAX_SAMPLES_DISPLAY
from letta_evals.datasets.loader import load_jsonl
from letta_evals.models import RunnerResult, SuiteSpec
from letta_evals.runner import run_suite
from letta_evals.visualization.factory import ProgressStyle

app = typer.Typer(help="Letta Evals - Evaluation framework for Letta AI agents")
console = Console()


@app.command()
def run(
    suite_path: Path = typer.Argument(..., help="Path to suite YAML file"),
    output: Optional[Path] = typer.Option(
        None, "--output", "-o", help="Stream header, summary, and per-instance results to directory"
    ),
    quiet: bool = typer.Option(False, "--quiet", "-q", help="Minimal output"),
    max_concurrent: int = typer.Option(15, "--max-concurrent", help="Maximum concurrent evaluations"),
    cached: Optional[Path] = typer.Option(
        None, "--cached", "-c", help="Path to cached results (JSONL) for re-grading trajectories"
    ),
    api_key: Optional[str] = typer.Option(
        None,
        "--api-key",
        help="Letta API key override. If not provided, uses LETTA_API_KEY from environment",
    ),
    base_url: Optional[str] = typer.Option(
        None,
        "--base-url",
        help="Letta base URL override. If omitted and an API key is set, defaults to Letta Cloud",
    ),
    project_id: Optional[str] = typer.Option(
        None,
        "--project-id",
        help="Letta project ID override. If not provided, uses LETTA_PROJECT_ID from environment or suite config",
    ),
):
    """Run an evaluation suite."""

    # auto-detect if we should disable fancy output based on terminal capabilities
    import os

    no_fancy = not console.is_terminal or os.getenv("NO_COLOR") is not None

    # verbose is now the default unless --quiet is specified
    verbose = not quiet

    if not suite_path.exists():
        console.print(f"[red]Error: Suite file not found: {suite_path}[/red]")
        raise typer.Exit(1)

    try:
        with open(suite_path, "r") as f:
            yaml_data = yaml.safe_load(f)
        suite = SuiteSpec.from_yaml(yaml_data, base_dir=suite_path.parent)

        samples = list(load_jsonl(suite.dataset, max_samples=suite.max_samples, sample_tags=suite.sample_tags))
        num_samples = len(samples)

        # calculate total evaluations (samples × models)
        if suite.target.model_configs:
            num_models = len(suite.target.model_configs)
        elif suite.target.model_handles:
            num_models = len(suite.target.model_handles)
        else:
            num_models = 1
        total_evaluations = num_samples * num_models
    except Exception as e:
        console.print(f"[red]Error loading suite: {e}[/red]")
        raise typer.Exit(1)

    if not quiet and not no_fancy:
        console.print(f"[cyan]Loading suite: {suite.name}[/cyan]")
        if num_models > 1:
            console.print(
                f"[cyan]Total evaluations: {total_evaluations} ({num_samples} samples × {num_models} models)[/cyan]"
            )
        else:
            console.print(f"[cyan]Total samples: {num_samples}[/cyan]")
        console.print(f"[cyan]Max concurrent: {max_concurrent}[/cyan]")

        if cached:
            console.print(f"[yellow]Using cached trajectories from: {cached}[/yellow]")
            console.print(
                f"[yellow]Re-grading {total_evaluations} trajectories with updated grader configuration[/yellow]"
            )

    async def run_with_progress():
        # Choose built-in progress style for CLI
        if quiet:
            style = ProgressStyle.NONE
        elif no_fancy:
            style = ProgressStyle.SIMPLE
        else:
            style = ProgressStyle.RICH

        if not quiet:
            console.print(f"Running evaluation suite: {suite.name}")
            if cached:
                console.print(f"[yellow]Re-grading {total_evaluations} cached trajectories...[/yellow]")
            else:
                console.print(f"Evaluating {total_evaluations} samples...")

        return await run_suite(
            suite_path,
            max_concurrent=max_concurrent,
            progress_style=style,
            cached_results_path=cached,
            output_path=output,
            letta_api_key=api_key,
            letta_base_url=base_url,
            letta_project_id=project_id,
        )

    try:
        result = anyio.run(run_with_progress)  # type: ignore[arg-type]

        if not quiet:
            display_results(result, verbose, cached_mode=(cached is not None))

        if output and not quiet:
            console.print(f"[green]Results streamed to {output}/results.jsonl (JSONL)[/green]")
            console.print(f"[green]Summary saved to {output}/summary.json[/green]")
            console.print(f"[green]Header saved to {output}/header.json[/green]")

        if result.gates_passed:
            if not quiet:
                console.print("[green]✓ All gates passed[/green]")
            sys.exit(0)
        else:
            if not quiet:
                console.print("[red]✗ Some gates failed[/red]")
            sys.exit(1)

    except Exception as e:
        console.print(f"[red]Error running suite: {e}[/red]")
        if verbose:
            import traceback

            traceback.print_exc()
        raise typer.Exit(1)


@app.command()
def validate(suite_path: Path = typer.Argument(..., help="Path to suite YAML file")):
    """Validate a suite configuration without running it."""

    if not suite_path.exists():
        console.print(f"[red]Error: Suite file not found: {suite_path}[/red]")
        raise typer.Exit(1)

    try:
        with open(suite_path, "r") as f:
            yaml_data = yaml.safe_load(f)

        suite = SuiteSpec.from_yaml(yaml_data, base_dir=suite_path.parent)
        console.print(f"[green]✓ Suite '{suite.name}' is valid[/green]")

        console.print("\n[bold]Configuration:[/bold]")
        console.print(f"  Dataset: {suite.dataset}")
        console.print(f"  Target: {suite.target.kind.value}")
        if suite.graders:
            console.print("  Graders:")
            for key, gspec in suite.graders.items():
                label = gspec.display_name or key
                console.print(f"    - {label}: {gspec.kind.value}")
        if suite.gate:
            metric_key = suite.gate.metric_key or "<default>"
            console.print(
                f"  Gate: metric_key={metric_key} aggregate={suite.gate.metric.value} {suite.gate.op.value} {suite.gate.value}"
            )

    except Exception as e:
        console.print(f"[red]Invalid suite configuration: {e}[/red]")
        raise typer.Exit(1)


@app.command("list-extractors")
def list_extractors():
    """List available submission extractors."""

    from letta_evals.decorators import EXTRACTOR_REGISTRY

    table = Table(title="Available Extractors")
    table.add_column("Name", style="cyan")
    table.add_column("Description", style="white")

    descriptions = {
        "last_assistant": "Extract the last assistant message",
        "first_assistant": "Extract the first assistant message",
        "all_assistant": "Concatenate all assistant messages",
        "last_turn": "Extract assistant messages from last turn",
        "pattern": "Extract using regex pattern",
        "json": "Extract JSON field from response",
        "tool_output": "Extract specific tool output",
        "after_marker": "Extract content after marker",
    }

    for name in sorted(EXTRACTOR_REGISTRY.keys()):
        desc = descriptions.get(name, "")
        table.add_row(name, desc)

    console.print(table)


@app.command("list-graders")
def list_graders():
    """List available built-in grader functions."""

    from letta_evals.graders.tool import GRADER_REGISTRY

    table = Table(title="Built-in Graders")
    table.add_column("Name", style="cyan")
    table.add_column("Type", style="yellow")

    for name in sorted(GRADER_REGISTRY.keys()):
        table.add_row(name, "tool")

    console.print(table)
    console.print("\n[dim]You can also use 'rubric' graders with custom prompts[/dim]")


def display_results(result: RunnerResult, verbose: bool = False, cached_mode: bool = False):
    console.print(f"\n[bold]Evaluation Results: {result.suite}[/bold]")
    if cached_mode:
        console.print("[dim]Note: Results re-graded from cached trajectories[/dim]")
    console.print("=" * 50)

    metrics = result.metrics
    console.print("\n[bold]Overall Metrics:[/bold]")
    console.print(f"  Total samples: {metrics.total}")
    console.print(f"  Total attempted: {metrics.total_attempted}")
    errors = metrics.total - metrics.total_attempted
    errors_pct = (errors / metrics.total * 100.0) if metrics.total > 0 else 0.0
    console.print(f"  Errored: {errors_pct:.1f}% ({errors}/{metrics.total})")
    console.print(f"  Average score (attempted, gate metric): {metrics.avg_score_attempted:.2f}")
    console.print(f"  Average score (total, gate metric): {metrics.avg_score_total:.2f}")
    console.print(f"  Passed attempts (gate metric): {metrics.passed_attempts}")
    console.print(f"  Failed attempts (gate metric): {metrics.failed_attempts}")

    # Print per-metric aggregates if available
    if hasattr(metrics, "by_metric") and metrics.by_metric:
        console.print("\n[bold]Metrics by Metric:[/bold]")
        table = Table()
        table.add_column("Metric", style="cyan")
        table.add_column("Avg Score (Attempted)", style="white")
        table.add_column("Avg Score (Total)", style="white")
        # Build key->label mapping from config
        label_map = {}
        if "graders" in result.config and isinstance(result.config["graders"], dict):
            for key, gspec in result.config["graders"].items():
                label_map[key] = gspec.get("display_name") or key

        for key, agg in metrics.by_metric.items():
            label = label_map.get(key, key)
            table.add_row(label, f"{agg.avg_score_attempted:.2f}", f"{agg.avg_score_total:.2f}")
        console.print(table)

    # show per-model metrics if available
    if metrics.per_model:
        console.print("\n[bold]Per-Model Metrics:[/bold]")
        model_table = Table()
        model_table.add_column("Model", style="cyan")
        model_table.add_column("Samples", style="white")
        model_table.add_column("Attempted", style="white")
        model_table.add_column("Avg Score (Attempted)", style="white")
        model_table.add_column("Avg Score (Total)", style="white")
        model_table.add_column("Passed", style="green")
        model_table.add_column("Failed", style="red")

        for model_metrics in metrics.per_model:
            model_table.add_row(
                model_metrics.model_name,
                str(model_metrics.total),
                str(model_metrics.total_attempted),
                f"{model_metrics.avg_score_attempted:.2f}",
                f"{model_metrics.avg_score_total:.2f}",
                str(model_metrics.passed_samples),
                str(model_metrics.failed_samples),
            )

        console.print(model_table)

    gate = result.config["gate"]
    gate_op = gate["op"]
    gate_value = gate["value"]
    gate_metric = gate.get("metric", "avg_score")
    gate_metric_key = gate.get("metric_key")

    op_symbols = {"gt": ">", "gte": "≥", "lt": "<", "lte": "≤", "eq": "="}
    op_symbol = op_symbols.get(gate_op, gate_op)

    status = "[green]PASSED[/green]" if result.gates_passed else "[red]FAILED[/red]"

    if gate_metric == "avg_score":
        actual = metrics.avg_score_attempted
        suffix = ""
    else:
        if gate_metric_key and gate_metric_key in metrics.metrics:
            actual = metrics.metrics[gate_metric_key]
        elif metrics.metrics:
            actual = next(iter(metrics.metrics.values()))
        else:
            actual = 0.0
        suffix = "%"

    # Prefer display name for gate metric key
    display_label = None
    if gate_metric_key and "graders" in result.config and isinstance(result.config["graders"], dict):
        gspec = result.config["graders"].get(gate_metric_key)
        if gspec:
            display_label = gspec.get("display_name")
    metric_key_suffix = f" on '{display_label or gate_metric_key}'" if gate_metric_key else ""
    console.print(
        f"\n[bold]Gate:{metric_key_suffix}[/bold] {gate_metric} {op_symbol} {gate_value:.2f}{suffix} → {status} (actual: {actual:.2f}{suffix}, total: {metrics.avg_score_total:.2f})"
    )

    if verbose:
        total_samples = len(result.results)
        samples_to_display = result.results[:MAX_SAMPLES_DISPLAY]

        console.print("\n[bold]Sample Results:[/bold]")
        if total_samples > MAX_SAMPLES_DISPLAY:
            console.print(f"[dim]Showing first {MAX_SAMPLES_DISPLAY} of {total_samples} samples[/dim]")

        table = Table()
        table.add_column("Sample", style="cyan")
        table.add_column("Model", style="yellow")
        table.add_column("Passed", style="white")

        # Determine available metrics and display labels
        metric_keys = []
        metric_labels = {}
        if "graders" in result.config and isinstance(result.config["graders"], dict):
            for k, gspec in result.config["graders"].items():
                metric_keys.append(k)
                metric_labels[k] = gspec.get("display_name") or k

        # Add two sub-columns per metric: score + rationale
        for mk in metric_keys:
            lbl = metric_labels.get(mk, mk)
            table.add_column(f"{lbl} score", style="white")
            table.add_column(f"{lbl} rationale", style="dim")

        from letta_evals.models import GateSpec

        gate_spec = GateSpec(**result.config["gate"])

        for sample_result in samples_to_display:
            score_val = sample_result.grade.score
            passed = "✓" if gate_spec.check_sample(score_val) else "✗"

            # Build per-metric cells in config order
            cells = []
            for mk in metric_keys:
                g = sample_result.grades.get(mk) if sample_result.grades else None
                if g is None:
                    cells.extend(["-", ""])
                else:
                    try:
                        s_val = float(getattr(g, "score", None))
                        r_text = getattr(g, "rationale", None) or ""
                    except Exception:
                        # dict fallback
                        try:
                            s_val = float(g.get("score"))  # type: ignore[attr-defined]
                            r_text = g.get("rationale", "")  # type: ignore[attr-defined]
                        except Exception:
                            s_val = None
                            r_text = ""
                    score_cell = f"{s_val:.2f}" if s_val is not None else "-"
                    if r_text and len(r_text) > 50:
                        r_text = r_text[:47] + "..."
                    cells.extend([score_cell, r_text])

            table.add_row(f"Sample {sample_result.sample.id + 1}", sample_result.model_name or "-", passed, *cells)

        console.print(table)

        if total_samples > MAX_SAMPLES_DISPLAY:
            console.print(
                f"[dim]... and {total_samples - MAX_SAMPLES_DISPLAY} more samples (see output file for complete results)[/dim]"
            )


if __name__ == "__main__":
    app()



import inspect
from functools import wraps
from typing import Callable, Dict

from letta_evals.models import GradeResult

GRADER_REGISTRY: Dict[str, Callable] = {}
EXTRACTOR_REGISTRY: Dict[str, Callable] = {}


def grader(func: Callable = None, *, name: str = None):
    """
    Decorator for grader functions.

    Validates that the function has signature: (Sample, str) -> GradeResult
    Auto-registers the function to the grader registry.

    Usage:
        @grader
        def my_grader(sample: Sample, submission: str) -> GradeResult:
            ...

        @grader(name="custom_name")
        def another_grader(sample: Sample, submission: str) -> GradeResult:
            ...
    """

    def decorator(f: Callable) -> Callable:
        sig = inspect.signature(f)
        params = list(sig.parameters.values())

        if len(params) != 2:
            raise TypeError(
                f"Grader {f.__name__} must have exactly 2 parameters (sample: Sample, submission: str), "
                f"got {len(params)}"
            )

        param_names = [p.name for p in params]
        if param_names != ["sample", "submission"]:
            raise TypeError(
                f"Grader {f.__name__} must have parameters named 'sample' and 'submission', got {param_names}"
            )

        if sig.return_annotation != inspect.Signature.empty:
            if sig.return_annotation != GradeResult:
                raise TypeError(f"Grader {f.__name__} must return GradeResult, got {sig.return_annotation}")

        registry_name = name or f.__name__
        GRADER_REGISTRY[registry_name] = f

        f._is_grader = True

        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapper

    if func is None:
        return decorator
    else:
        return decorator(func)


def extractor(func: Callable = None, *, name: str = None):
    """
    Decorator for extractor functions.

    Validates that the function has signature:
      - (trajectory: List[List[LettaMessageUnion]], config: dict) -> str, OR
      - (trajectory: List[List[LettaMessageUnion]], config: dict, agent_state: Optional[AgentState]) -> str
    Auto-registers the function to the extractor registry.

    Usage:
        @extractor
        def my_extractor(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
            ...

        @extractor
        def memory_extractor(trajectory: List[List[LettaMessageUnion]], config: dict, agent_state: Optional[AgentState]) -> str:
            ...

        @extractor(name="custom_name")
        def another_extractor(trajectory: List[List[LettaMessageUnion]], config: dict) -> str:
            ...
    """

    def decorator(f: Callable) -> Callable:
        sig = inspect.signature(f)
        params = list(sig.parameters.values())

        if len(params) not in (2, 3):
            raise TypeError(
                f"Extractor {f.__name__} must have 2 or 3 parameters (trajectory, config) or (trajectory, config, agent_state), got {len(params)}"
            )

        param_names = [p.name for p in params]
        if len(params) == 2:
            if param_names != ["trajectory", "config"]:
                raise TypeError(
                    f"Extractor {f.__name__} must have parameters named 'trajectory' and 'config', got {param_names}"
                )
        elif len(params) == 3:
            if param_names != ["trajectory", "config", "agent_state"]:
                raise TypeError(
                    f"Extractor {f.__name__} must have parameters named 'trajectory', 'config', and 'agent_state', got {param_names}"
                )

        if sig.return_annotation != inspect.Signature.empty:
            if sig.return_annotation is not str:
                raise TypeError(f"Extractor {f.__name__} must return str, got {sig.return_annotation}")

        registry_name = name or f.__name__
        EXTRACTOR_REGISTRY[registry_name] = f

        f._is_extractor = True
        f._extractor_param_count = len(params)

        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapper

    if func is None:
        return decorator
    else:
        return decorator(func)


def agent_factory(func: Callable) -> Callable:
    """
    Decorator for agent factory functions.

    Validates that the function has signature: async (client: AsyncLetta, sample: Sample) -> str

    Usage:
        @agent_factory
        async def create_inventory_agent(client: AsyncLetta, sample: Sample) -> str:
            # create agent using client and sample data
            return agent_id
    """
    sig = inspect.signature(func)
    params = list(sig.parameters.values())

    if len(params) != 2:
        raise TypeError(
            f"Agent factory {func.__name__} must have exactly 2 parameters (client, sample), got {len(params)}"
        )

    param_names = [p.name for p in params]
    if param_names != ["client", "sample"]:
        raise TypeError(
            f"Agent factory {func.__name__} must have parameters named 'client' and 'sample', got {param_names}"
        )

    if not inspect.iscoroutinefunction(func):
        raise TypeError(f"Agent factory {func.__name__} must be an async function")

    if sig.return_annotation != inspect.Signature.empty:
        if sig.return_annotation is not str:
            raise TypeError(f"Agent factory {func.__name__} must return str (agent_id), got {sig.return_annotation}")

    # mark as validated agent factory
    func._is_agent_factory = True

    @wraps(func)
    async def wrapper(*args, **kwargs):
        return await func(*args, **kwargs)

    return wrapper


def suite_setup(func: Callable) -> Callable:
    """
    Decorator for suite setup functions.

    Validates that the function has signature: async (client: AsyncLetta) -> None
    Also supports sync functions: (client: AsyncLetta) -> None

    Usage:
        @suite_setup
        async def prepare_evaluation(client: AsyncLetta) -> None:
            # perform setup operations with client
            await client.tools.add(tool=MyCustomTool())

        @suite_setup
        def prepare_evaluation_sync(client: AsyncLetta) -> None:
            # perform sync setup operations
            pass
    """
    sig = inspect.signature(func)
    params = list(sig.parameters.values())

    if len(params) != 1:
        raise TypeError(f"Suite setup {func.__name__} must have exactly 1 parameter (client), got {len(params)}")

    param_names = [p.name for p in params]
    if param_names != ["client"]:
        raise TypeError(f"Suite setup {func.__name__} must have parameter named 'client', got {param_names}")

    if sig.return_annotation != inspect.Signature.empty:
        if sig.return_annotation is not None and sig.return_annotation is not None:
            raise TypeError(f"Suite setup {func.__name__} must return None, got {sig.return_annotation}")

    # mark as validated suite setup
    func._is_suite_setup = True

    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def wrapper(*args, **kwargs):
            return await func(*args, **kwargs)
    else:

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

    return wrapper



import importlib.util
import sys
from pathlib import Path
from typing import Any


def load_object(spec: str, base_dir: Path = None) -> Any:
    """Load a Python object from a file path specification."""
    if not spec:
        raise ValueError("Empty specification provided")

    if ":" not in spec:
        raise ImportError(f"'{spec}' appears to be a simple name, not a file path")

    file_path, obj_name = spec.rsplit(":", 1)
    path = Path(file_path)

    # resolve relative paths
    if not path.is_absolute():
        if base_dir is None:
            raise ValueError(f"Relative path provided but no base_dir: {file_path}")
        path = (base_dir / path).resolve()

    if not path.exists():
        raise FileNotFoundError(f"File not found: {path}")

    if path.suffix != ".py":
        raise ValueError(f"File must be a Python file (.py), got: {path}")

    module_name = f"_dynamic_{path.stem}_{id(path)}"
    spec_loader = importlib.util.spec_from_file_location(module_name, path)
    if spec_loader is None or spec_loader.loader is None:
        raise ImportError(f"Could not load module from {path}")

    module = importlib.util.module_from_spec(spec_loader)
    sys.modules[module_name] = module
    spec_loader.loader.exec_module(module)

    if not hasattr(module, obj_name):
        available = [name for name in dir(module) if not name.startswith("_")]
        raise AttributeError(f"Module '{path}' has no attribute '{obj_name}'. Available: {', '.join(available[:10])}")

    return getattr(module, obj_name)



from __future__ import annotations

from typing import Dict, Optional

from rich.console import Console

from letta_evals.visualization.base import ProgressCallback


class SimpleProgress(ProgressCallback):
    """Clean hierarchical progress callback for CI and non-interactive terminals.

    Uses visual hierarchy with indentation and simple unicode symbols to make
    evaluation progress easy to scan in logs.
    """

    def __init__(self, suite_name: str, total_samples: int, console: Optional[Console] = None):
        self.suite_name = suite_name
        self.total_samples = total_samples
        self.console = console or Console()
        self._current_sample = None

    async def start(self) -> None:
        self.console.print("━" * 60)
        self.console.print(f"[bold cyan]Suite:[/] {self.suite_name}")
        self.console.print(f"[bold cyan]Samples:[/] {self.total_samples}")
        self.console.print("━" * 60)
        self.console.print()

    def stop(self) -> None:
        self.console.print()
        self.console.print("━" * 60)
        self.console.print("[bold cyan]Suite completed[/]")
        self.console.print("━" * 60)

    async def sample_started(self, sample_id: int, model_name: Optional[str] = None) -> None:
        # track current sample to avoid printing header multiple times
        self._current_sample = (sample_id, model_name)
        model_text = f" [dim]({model_name})[/]" if model_name else ""
        self.console.print(f"[bold cyan]▸ Sample [{sample_id}]{model_text}[/]")

    async def agent_loading(self, sample_id: int, model_name: Optional[str] = None, from_cache: bool = False) -> None:
        prefix = self._format_prefix(sample_id, model_name)
        cache_text = " [dim](cached)[/]" if from_cache else ""
        self.console.print(f"{prefix} [dim]•[/] Loading agent{cache_text}")

    async def message_sending(
        self, sample_id: int, message_num: int, total_messages: int, model_name: Optional[str] = None
    ) -> None:
        prefix = self._format_prefix(sample_id, model_name)
        self.console.print(f"{prefix} [dim]•[/] Sending messages {message_num}/{total_messages}")

    async def grading_started(self, sample_id: int, model_name: Optional[str] = None) -> None:
        prefix = self._format_prefix(sample_id, model_name)
        self.console.print(f"{prefix} [dim]•[/] Grading...")

    async def sample_completed(
        self,
        sample_id: int,
        passed: bool,
        score: Optional[float] = None,
        model_name: Optional[str] = None,
        metric_scores: Optional[Dict[str, float]] = None,
        metric_pass: Optional[Dict[str, bool]] = None,
        rationale: Optional[str] = None,
        metric_rationales: Optional[Dict[str, str]] = None,
    ) -> None:
        prefix = self._format_prefix(sample_id, model_name)
        status = "[bold green]✓ PASS[/]" if passed else "[bold red]✗ FAIL[/]"
        parts = [f"{prefix} {status}"]

        if score is not None:
            parts.append(f"score={score:.2f}")

        if metric_scores:
            metric_bits = ", ".join(f"{k}={v:.2f}" for k, v in metric_scores.items())
            parts.append(metric_bits)

        self.console.print("  ".join(parts))

    async def sample_error(self, sample_id: int, error: str, model_name: Optional[str] = None) -> None:
        prefix = self._format_prefix(sample_id, model_name)
        self.console.print(f"{prefix} [bold yellow]⚠ ERROR[/]: {error}")

    def _format_prefix(self, sample_id: int, model_name: Optional[str]) -> str:
        """format a compact prefix for substeps to show which sample they belong to."""
        if model_name:
            return f"[dim]\\[[/][cyan]{sample_id}[/][dim]]({model_name})[/]"
        return f"[dim]\\[[/][cyan]{sample_id}[/][dim]][/]"



# import builtin extractors to register them
from letta_evals.extractors.builtin import (
    after_marker,
    all_assistant,
    first_assistant,
    last_assistant,
    last_turn,
    memory_block,
    pattern,
    tool_arguments,
    tool_output,
)
from letta_evals.extractors.registry import extractor_requires_agent_state, get_extractor

__all__ = [
    "after_marker",
    "all_assistant",
    "first_assistant",
    "last_assistant",
    "last_turn",
    "memory_block",
    "pattern",
    "tool_arguments",
    "tool_output",
    "get_extractor",
    "extractor_requires_agent_state",
]



from pathlib import Path
from typing import List, Optional, Tuple

from letta_client import AgentState, LettaMessageUnion

from letta_evals.decorators import GRADER_REGISTRY
from letta_evals.extractors import extractor_requires_agent_state, get_extractor
from letta_evals.graders.base import Grader
from letta_evals.models import GradeResult, Sample
from letta_evals.utils import load_object


class ToolGrader(Grader):
    """Grader that uses Python functions."""

    def __init__(
        self,
        function: str,
        extractor: str = "last_assistant",
        extractor_config: Optional[dict] = None,
        base_dir: Optional[Path] = None,
    ):
        self.function_name = function
        self.extractor_name = extractor
        self.base_dir = base_dir
        self.extractor = get_extractor(extractor, extractor_config, base_dir=base_dir)
        self._requires_agent_state = extractor_requires_agent_state(extractor, base_dir=base_dir)

        if function in GRADER_REGISTRY:
            self.func = GRADER_REGISTRY[function]
        elif ":" in function:
            obj = load_object(function, base_dir=base_dir)
            if callable(obj) and hasattr(obj, "_is_grader"):
                self.func = obj
            else:
                raise ValueError(
                    f"Loaded object {function} is not a valid @grader decorated function. "
                    f"Please use the @grader decorator."
                )
        else:
            raise ValueError(f"Grader function '{function}' not found in registry")

    @property
    def requires_agent_state(self) -> bool:
        """Whether this grader's extractor requires agent_state."""
        return self._requires_agent_state

    async def grade(
        self, sample: Sample, trajectory: List[List[LettaMessageUnion]], agent_state: Optional[AgentState] = None
    ) -> Tuple[GradeResult, str]:
        """Grade using the tool function."""
        submission = self.extractor(trajectory, agent_state=agent_state)
        return self.func(sample, submission), submission



from __future__ import annotations

from enum import Enum
from typing import Dict, Optional

from rich.console import Console

from letta_evals.models import SuiteSpec
from letta_evals.visualization.base import ProgressCallback
from letta_evals.visualization.noop_progress import NoOpProgress
from letta_evals.visualization.rich_progress import DisplayMode, EvalProgress
from letta_evals.visualization.simple_progress import SimpleProgress


class ProgressStyle(str, Enum):
    """Built-in progress verbosity levels."""

    NONE = "none"
    SIMPLE = "simple"
    RICH = "rich"


def create_progress_callback(
    style: ProgressStyle,
    suite: SuiteSpec,
    total_evaluations: int,
    *,
    console: Optional[Console] = None,
    max_concurrent: int = 15,
    cached_mode: bool = False,
    metric_labels: Optional[Dict[str, str]] = None,
) -> ProgressCallback:
    """Factory for built-in progress callbacks.

    - RICH: full Rich-based UI from visualization.progress
    - SIMPLE: single-line updates suitable for logs
    - NONE: silent
    """
    if style == ProgressStyle.NONE:
        return NoOpProgress()

    if style == ProgressStyle.SIMPLE:
        return SimpleProgress(suite_name=suite.name, total_samples=total_evaluations, console=console)

    # RICH (default for CLI)
    # Determine grader kind label for header
    if suite.graders and len(suite.graders) == 1:
        only_kind = next(iter(suite.graders.values())).kind.value
        grader_kind_label = only_kind
    else:
        grader_kind_label = "multi"

    # Choose rubric model if any grader is rubric
    rubric_model = None
    if suite.graders:
        for _, gspec in suite.graders.items():
            if hasattr(gspec, "kind") and getattr(gspec, "kind").value == "rubric" and hasattr(gspec, "model"):
                rubric_model = gspec.model
                break

    progress = EvalProgress(
        suite_name=suite.name,
        total_samples=total_evaluations,
        target_kind=suite.target.kind.value,
        grader_kind=grader_kind_label,
        rubric_model=rubric_model,
        max_concurrent=max_concurrent,
        display_mode=DisplayMode.DETAILED,
        console=console,
        show_samples=True,
        cached_mode=cached_mode,
        metric_labels=metric_labels,
    )
    return progress


__all__ = ["ProgressStyle", "create_progress_callback"]



from typing import Any, List

from letta_client import (
    AssistantMessage,
    LettaMessageUnion,
    ToolCallMessage,
    ToolReturnMessage,
)


def flatten_content(content: Any) -> str:
    """Flatten message content to string."""
    if isinstance(content, str):
        return content
    elif isinstance(content, list):
        parts = [item.text for item in content]
        return " ".join(parts)
    else:
        raise ValueError(f"Unexpected content type: {type(content)}")


def get_assistant_messages(trajectory: List[List[LettaMessageUnion]]) -> List[AssistantMessage]:
    """Extract all assistant messages from trajectory."""
    messages = []
    for turn in trajectory:
        for msg in turn:
            if isinstance(msg, AssistantMessage):
                messages.append(msg)
    return messages


def get_tool_calls(trajectory: List[List[LettaMessageUnion]]) -> List[tuple[ToolCallMessage, ToolReturnMessage]]:
    """Extract all tool call/return pairs from trajectory."""
    pairs = []
    for turn in trajectory:
        i = 0
        while i < len(turn):
            if isinstance(turn[i], ToolCallMessage):
                for j in range(i + 1, len(turn)):
                    if isinstance(turn[j], ToolReturnMessage):
                        pairs.append((turn[i], turn[j]))
                        break
            i += 1
    return pairs


def get_messages_by_type(trajectory: List[List[LettaMessageUnion]], message_type: type) -> List[LettaMessageUnion]:
    """Filter messages by type."""
    messages = []
    for turn in trajectory:
        for msg in turn:
            if isinstance(msg, message_type):
                messages.append(msg)
    return messages


def get_last_turn_messages(
    trajectory: List[List[LettaMessageUnion]], message_type: type = AssistantMessage
) -> List[LettaMessageUnion]:
    """Get all messages of a given type from the last turn."""
    if not trajectory:
        return []

    last_turn = trajectory[-1]
    return [msg for msg in last_turn if isinstance(msg, message_type)]



"""Letta Evals Kit - Evaluation framework for Letta AI agents."""

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _pkg_version

from letta_evals.graders.rubric import RubricGrader
from letta_evals.graders.tool import ToolGrader
from letta_evals.models import GateSpec, RunnerResult, Sample, SuiteSpec
from letta_evals.runner import Runner, run_suite
from letta_evals.targets.agent import AgentTarget

try:
    __version__: str = _pkg_version("letta-evals")
except PackageNotFoundError:
    __version__ = "0.3.2"

__all__ = [
    "Sample",
    "SuiteSpec",
    "RunnerResult",
    "GateSpec",
    "Runner",
    "run_suite",
    "AgentTarget",
    "ToolGrader",
    "RubricGrader",
]



from abc import ABC, abstractmethod
from typing import Optional

from letta_evals.models import Sample, TargetResult
from letta_evals.visualization.base import ProgressCallback


class Target(ABC):
    """Base interface for evaluation targets."""

    @abstractmethod
    async def run(self, sample: Sample, progress_callback: Optional[ProgressCallback] = None) -> TargetResult:
        """Run the target on a sample and return result."""
        pass



import json
from pathlib import Path
from typing import Iterator, List, Optional

from letta_evals.models import Sample


def load_jsonl(
    file_path: Path, max_samples: Optional[int] = None, sample_tags: Optional[List[str]] = None
) -> Iterator[Sample]:
    """Load samples from a JSONL file."""
    with open(file_path, "r") as f:
        line_index = 0
        yielded_count = 0
        for line in f:
            if max_samples and yielded_count >= max_samples:
                break

            data = json.loads(line.strip())

            # skip filtering by tags since metadata is removed
            if sample_tags:
                # tags filtering no longer supported without metadata
                pass

            sample = Sample(
                id=line_index,
                input=data["input"],
                ground_truth=data.get("ground_truth"),
                agent_args=data.get("agent_args"),
            )

            line_index += 1
            yielded_count += 1
            yield sample
